// Eigenstates of the 1D Schroedinger equation for the harmonic oscillator
//    y" = 2 [V(x) - E] y, y(0) = y0, y'(0) = y'0, y(+inf) = 0
//    V(x) = 0.5*x*x
// using a shooting algorithm based on the Numerov method
//---------------------------------------------------------------------------
#include "memalloc.h"
#include "ode.h"
#include "specfunc.h"

#define pi 3.141592653589793
                                         // Potential for harmonic oscillator
double Pot(double x) { return 0.5e0*x*x; } 

int main()
{
   double *V, *x, *y;
   double d, dE, dy0, E, Emin, Emax, Ew, eps, f, hx, xc, xx, y0, yH;
   int exist, i, m, n, nc, nx, par;
   FILE *out;

   xx = 10e0;                                              // limit of x-mesh
   xc = 6e0;                   // checkpoint for vanishing solution (xc < xx)
   par = 0;                                     // parity of eigenstate (0/1)
   Emin = 0e0;                                 // lower limit for eigenvalues
   Emax = 1e2;                                 // upper limit for eigenvalues
   dE = 0.1e0;                           // minimum separation of eigenvalues
   eps = 1e-4;                                // tolerance for solution at xc
   hx = 1e-4;                                             // x-mesh step size

   nx = int(xx/hx + 0.5) + 1;                      // number of x-mesh points
   nc = int(xc/hx + 0.5) + 1;          // index of checkpoint for vanishing y
   x = Vector(1,nx); y = Vector(1,nx);                    // x-mesh, solution
   V = Vector(1,nx);                                   // tabulated potential

   for (m=1; m<=nx; m++) {
      x[m] = (m-1)*hx;                                    // integration mesh
      V[m] = Pot(x[m]);                                 // tabulate potential
   }

   Ew = Emin;               // lower limit of search window for E, [Ew,Ew+dE]
   while (Ew < Emax) {                        // loop over eigenvalue windows
                                                   // initial values at x = 0
      if (par == 0) { y0 = 1e0; dy0 = 0e0; }                        // even y
      else          { y0 = 0e0; dy0 = 1e0; }                         // odd y
                                                       // shoot in [Ew,Ew+dE]
      E = ShootQM(Ew,Ew+dE,V,x,y,nx,nc,y0,dy0,eps,exist);

      if (exist) break;

      Ew += dE;                            // shift [Ew,Ew+dE] for next shoot
   }

   if (exist) {
      n = int(E);                                           // quantum number

      f = 0e0;                             // normalize y by trapezoidal rule
      for (m=1; m<=nc; m++) f += y[m]*y[m]*hx;            // norm on [0,+inf]
      f = 1e0/sqrt(2e0*f);
      if (int(n/2) % 2) f = -f;                            // sign correction
      for (m=1; m<=nc; m++) y[m] *= f;

      f = sqrt(pi);                            // norm of Hermite polynomials
      for (i=1; i<=n; i++) f *= 2e0 * i;
      f = 1e0/sqrt(f);

      out = fopen("shoot.txt","w");
      fprintf(out,"E%i = %8.5f\n",n,E);
      fprintf(out,"      x        y%i        err\n",n);
      for (m=1; m<=nc; m++) {
         yH = f * exp(-0.5*x[m]*x[m]) * Hermite(n,x[m],d);
         fprintf(out,"%10.5f%10.5f%10.5f\n",x[m],y[m],yH-y[m]);
      }
      fclose(out);
   } else {
      printf("No solution found !\n");
   }
}
